/**
 * Class gérant les appel à l'API de AKS pour les jeux
 */

class AksApi {
    static url = null;

    /**
     * Initialise l'api en lui settant son endpoint
     * @param {string} url 
     */
    static init(url) {
        this.url = url;
    }

    static _ajaxCall(url) {
        return new Promise((resolve, reject) => {
            let xhttp = new XMLHttpRequest();
            xhttp.open('GET', url, true);
            xhttp.onreadystatechange = function() {
                if (this.readyState == 4 && this.status == 200) {
                    resolve(this.responseText);
                }
            };
            xhttp.send(); 
        });
    }

    static async call (params) {
        if (this.url === null) {
            throw 'API is not initialized';
        }
        let arrayParams = [];
        for (let name in params) {
            if (params[name] !== null) {
                arrayParams.push(name + '=' + params[name]);
            }
        }
        let urlParams = arrayParams.join('&');
        let url =  this.url + '?' + urlParams;
        let result = await this._ajaxCall(url)
        result = JSON.parse(result);
        return result;
    }

    /**
     * Execute une recherche de jeu sur leur nom
     * @param {string} text Le nom du jeu à chercher
     * @param {string} locale La locale à utiliser pour la recherche
     * @param {string} apiKey l'ID partenaire qui fait la recherche
     * @param {Object} filterParameters Filtres à ajouté pour la recherche
     *                                  - platform : N'afficher que les jeux d'une certaine console, valeurs acceptées : pc / playstation / xbox / nintendo
     *                                  - region : N'afficher que les jeux de la région souhaité (steam, epic, origin, ubisoft)
     * @returns 
     */
    static async searchGame(text, locale, apiKey, filterParameters) {
        filterParameters = (typeof(filterParameters) == 'undefined') ? {} : filterParameters;
        apiKey = (typeof(apiKey) === 'undefined') ? null : apiKey;
        let params = {
            action : 'products',
            showOffers : '0',
            locale : locale,
            currency : 'eur',
            apiKey : apiKey,
            search : encodeURIComponent(text)
        };
        let returnData = await this.call(params);
        if (returnData === null) return [];
        let returnGames = returnData.products;

        // Filtre sur la console de jeu
        if (filterParameters.hasOwnProperty('platform')) {
            let allowedConsoleRegexp = [new RegExp(filterParameters.platform, 'gi')];
            if (filterParameters.platform == 'nintendo') {
                allowedConsoleRegexp.push(new RegExp('switch', 'gi'));
                allowedConsoleRegexp.push(new RegExp('3ds', 'gi'));
            }
            returnGames = returnGames.filter((product) => {
                for(let i = 0, len = allowedConsoleRegexp.length; i < len; i++) {
                    if (allowedConsoleRegexp[i].test(product.platform)) {
                        return true
                    }
                }
                return false;
            });
        }

        // Filtre sur la région
        if (filterParameters.hasOwnProperty('region')) {
            let allowedRegion = new RegExp(filterParameters.region, 'gi');
            returnGames = returnGames.filter((product) => {
                if (product.bestOffer == null || !product.bestOffer.hasOwnProperty('region')) return false;
                return allowedRegion.test(product.bestOffer.region.id);
            });
        }
        return returnGames;
    }

    /**
     * Execute une recherche de jeu sur leur id
     * @param {array|number} ids le ou les ids à rechercher
     * @param {string} locale La locale à utiliser pour la recherche
     * @param {string} apiKey l'ID partenaire qui fait la recherche
     * @returns 
     */
        static async searchGameByIds(ids, locale, apiKey) {
            apiKey = (typeof(apiKey) === 'undefined') ? null : apiKey;
            if (!(ids instanceof Array)) {
                ids = [ids];
            }
            let params = {
                action : 'products',
                showOffers : '0',
                locale : locale,
                currency : 'usd',
                apiKey : apiKey,
                ids : encodeURIComponent(ids.join(','))
            };
            let returnData = await this.call(params);
            if (returnData === null) return [];
            return returnData.products;
        }

    /**
     * Retourne les données avec les offres d'un ou plusieurs jeux
     * @param {int|Array} games le ou les jeux
     * @param {string} locale la locale
     * @param {string} currency la devise
     * @param {string} apiKey l'api key
     * @return {Array}
     */
    static async getGamesData(games, locale, currency, apiKey) {
        apiKey = (typeof(apiKey) === 'undefined') ? null : apiKey;
        locale = (typeof(locale) === 'undefined') ? GLOBAL.locale : locale;
        currency = (typeof(currency) === 'undefined') ? GLOBAL.currency : currency;

        // Permet de vérifier si o na mis qu'un jeu aulieu d'un tableau de jeu en paramètre
        // afin de ne retourner que les données du jeu en question
        let oneGame = false;
        if (!(games instanceof Array)) {
            oneGame = true;
            games = [games];
        }

        let params = {
            action : 'products',
            showOffers : '1',
            locale : locale,
            currency : currency,
            apiKey : apiKey,
            ids : games.join(','),
            sortOffers : 'price-asc'
        };

        let returnData = await this.call(params);
        if (oneGame) {
            if(returnData === null) return null;
            if(returnData.products.length === 0) return null;
            return returnData.products[0];
        }
        if (returnData === null) return [];
        return returnData.products;
    }
}

AksApi.init('https://www.allkeyshop.com/api/latest/vaks.php');