<?php
namespace AllkeyshopAffiliate\Services;

class VaksApi
{

    const ENDPOINT = 'https://www.allkeyshop.com/api/latest/vaks.php';

    const MAX_API_RESULTS = 25;

    const FILTER_AVAILABLE = 1; // Filtre pour ne garder que les offres actuellement disponible
    const FILTER_SAME_EDITION = 2; // Filtre poru en garder que les offres ayant la même édition que celle du meilleur prix

    /**
     * Execute un appel à l'API
     * @param array $params les paramètre GET de l'appel
     * @param string $action le type d'appel à effectuer
     * @return array
     */
    public static function call(array $params = [], string $action = 'products')
    {
        // Condition spécial ici : si on recherche par ids, l'API est limité à 25 résultat maximum
        // On va donc couper le tableau des IDs en tranche de 25 et tout regrouper ensuite
        if (array_key_exists('ids', $params) && is_array($params['ids']) && count($params['ids']) > self::MAX_API_RESULTS) {
            $chunkIds = array_chunk($params['ids'], self::MAX_API_RESULTS);
            $allApiResult = false;
            foreach($chunkIds as $gameIds) {
                $apiParams = $params;
                $apiParams['ids'] = $gameIds;
                $apiResult = self::call($apiParams, $action);
                if ($apiResult === false) continue;

                if ($allApiResult === false) {
                    $allApiResult = $apiResult;
                } else {
                    $allApiResult['products'] = array_merge($allApiResult['products'], $apiResult['products']);
                }
            }
            return $allApiResult;
        }

        $stringParam = '?action=' . $action;
        if (!array_key_exists('currency', $params)) {
            $params['currency'] = 'usd';
        }
        foreach($params as $key => $value) {
            if (is_array($value)) {
                $value = implode(',', $value);
            } else if (is_bool($value)) {
                $value = ($value ? 1 : 0);
            }
            $stringParam .= '&' . $key . '=' . urlencode($value);
        }
        
        $response = wp_remote_retrieve_body(wp_remote_get(self::ENDPOINT . $stringParam));
        $decode = json_decode($response, true);
        if ($decode['status'] !== 'success') {
            // gestion des erreurs
            return false;
        }
        return $decode;
    }

    /**
     * Retourne toute les info des jeux en y incluant les offres
     * @param array $games un tableau d'id de jeu
     * @param string $locale
     * @param string $currency
     * @param string $apiKey l'ID partenaire
     * @param array $options un tableau de paramètres à envoyé à l'API
     *                       Les options de ce tableau sont prioritaire et écraserons les anciens paramètres
     * @return array le tableau ayant en clef l'id du jeu et en valeur ses infos
     */
    public static function getFullGamesData(array $games, string $locale, string $currency, string $apiKey = null, array $options = [], int $filters = 0)
    {
        if (empty($games)) {
            return [];
        }
        $params = [
            'ids' => $games,
            'locale' => $locale,
            'currency' => $currency,
            'showOffers' => 1
        ];
        if ($apiKey !== null) {
            $params['apiKey'] = $apiKey;
        }
        $fullParams = array_merge($params, $options);
        $data = self::call($fullParams);
        if ($data === false) return [];
        $gamesInformations = [];

        foreach($data['products'] as $value) {
            // On va appliquer les filtres sur chaque jeu
            if ($filters !== 0) {
                if (array_key_exists('offers', $value)) {
                    $filtersBool = [
                        'available' => (self::FILTER_AVAILABLE & $filters) == self::FILTER_AVAILABLE,
                        'sameEdition' => (self::FILTER_SAME_EDITION & $filters) == self::FILTER_SAME_EDITION
                    ];
                    $editionBestPrice = $value['bestOffer']['edition']['id'];
                    $value['offers'] = array_values(
                        array_filter(
                            $value['offers'],
                            function($offer) use ($filtersBool, $editionBestPrice) {
                                if ($filtersBool['available'] && $offer['available'] == false) {
                                    return false;
                                }
                                if ($filtersBool['sameEdition'] && $offer['edition']['id'] !== $editionBestPrice) {
                                    return false;
                                }
                                return true;
                            }
                        )
                    );
                }
            }

            // On y ajoute également des informations supplémentaire
            $value['platformFamily'] = self::_getPlatformFamily($value['platform']);
            $offers = [$value['bestOffer']];
            if (array_key_exists('offers', $value)) {
                $offers = $value['offers'];
            }
            $value['distributionPlatform'] = self::_getDistributionPlatform($offers);
            $gamesInformations[$value['id']] = $value;
        }
        return $gamesInformations;
    }

    /**
     * Returne le type de platforme (pc, playstation, xbox, nintendo) en fonction de la console du jeu
     * @param string $platform la console du jeu
     * @return string la famille de console
     */
    protected static function _getPlatformFamily(string $platform) 
    {
        if (preg_match('/xbox/i', $platform) === 1) return 'xbox';
        if (preg_match('/playstation/i', $platform) === 1) return 'playstation';
        if ($platform === 'switch' || $platform === '3ds') return 'nintendo';
        return 'pc';
    }

    /**
     * Returne la plateforme de distribution (steam, epic, origin, ubisoft) grâce aux offre d'un jeu
     * Il n'existe pas d'autre moyen de déterminer cette dernière actuellement
     * @param array $offers les offres du jeu
     * @return string la platforme
     */
    protected static function _getDistributionPlatform(array $offers) 
    {
        $platforms = ['steam', 'origin', 'epic', 'ubisoft'];
        foreach($offers as $offer) {
            if (array_key_exists('region', $offer)) {
                foreach($platforms as $platform) {
                    if (preg_match('/' . $platform . '/i', $offer['region']['id']) === 1) {
                        return $platform;
                    }
                }
            }
        }
        return null;
    }

    /**
     * Retourne toute les info des jeux sans y inclure les offres
     * @param array $games un tableau d'id de jeu
     * @param string $locale
     * @param string $currency
     * @param string $apiKey l'ID partenaire
     * @return array le tableau ayant en clef l'id du jeu et en valeur ses infos
     */
    public static function getGamesDataWithoutOffers(array $games, string $locale, string $currency, string $apiKey = null)
    {
        return self::getFullGamesData($games, $locale, $currency, $apiKey, ['showOffers' => 0]);
    }

    /**
     * Retourne le nom des jeux grâce à leur ids
     * @param int|array un ou plusieurs id de jeu
     * @return string|array Le nom du jeu si une seule valeur a été donné, un tableau avec les différents nom sinon
     */
    public static function getGamesTitle($games) {
        $sendOneResult = false;
        if (!is_array($games)) {
            $games = [$games];
            $sendOneResult = true;
        }
        $gameData = self::getFullGamesData($games, 'en_GB', 'eur', null, ['showOffers' => 0]);
        if ($sendOneResult) {
            $gameData = array_pop($gameData);
            return $gameData['name'];
        }
        $returnData = [];
        foreach($gameData as $data) {
            $returnData[$data['id']] = $data['name'];
        }
        return $returnData;
    }
}